/* =====================================================================
 * Copyright (c) 2003 Charles Gentry. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of this source code or a derived source code must
 *    retain the above copyright notice, this list of conditions and the
 *    following disclaimer.
 *
 * 2. Redistributions of this module or a derived module in binary form
 *    must reproduce the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY CHARLES GENTRY ``AS IS'' AND
 * ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL CHARLES GENTRY
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software is a contribution to and makes use of the Apache HTTP
 * server which is written, maintained and copywritten by the Apache Group.
 * See http://www.apache.org/ for more information.
 *
 * This software makes use of portions of the Apache server source code.
 *
 */
/*
 * mod_auth_msfix: Fix up Microsoft Webdav names
 *
 * Version:	0.2.1  2003 November
 * Author:	Charles Gentry
 *
 * This small module is just here to fix an MS XP problem where
 * it decides to prefix/postfix additional information onto
 * usernames when logging in to a webdav share.
 *
 * This program will take the input header from the client, determine
 * if it comes from a mangler, get rid of the extra information and then
 * put the header back.
 *
 * We are doing this during the AUTH phase rather than the fix-up headers
 * because we only care about this during the auth phase. Was a toss up 
 * between the two.
 *
 * A HUGE portion of this code was taken from the core module
 * 'protocol.c', subroutine 'ap_get_basic_pw'. We need to replace
 * the mangled line from MS XP clients (some) and just use the
 * standard user name.
 *
 * Some code was taken from the sample code supplied with
 * _Apache Modules_ by Stein and MacEachern.  
 *
 * ====================================================================
 */

#include "apr_strings.h"
#include "ap_config.h"

#include "httpd.h"  
#include "http_config.h"  
#include "http_core.h"  
#include "http_log.h"  
#include "http_protocol.h"  
#include "http_request.h"  

#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>


/* ====================================================================
 * Static constants
 */
static const char module_name[] = "auth_msfix_module";

/*
 * Function Prototypes
 */
static void *       auth_msfix_dir_config(apr_pool_t *p, char *d);
static int          auth_msfix_handler(request_rec *r);	/* Authentication */
static void         my_register_hooks();

static const char * add_msfix_rewrites( cmd_parms *, void *, 
			const char *, const char *);
static const char * add_msfix_onheader(  cmd_parms *, void *, 
			const char *, const char *);

/*
 * Structure definitions
 */
typedef struct {
	const char	*match;
	ap_regex_t	*compiled;
	const char	*replace;
} rewrite_entry;                                        // rewrite_entry

typedef struct {
	const char	*header;
	const char	*match;
	ap_regex_t	*compiled;
} matches_entry;                                        // matches_entry

typedef struct {
	int 			auth_msfix_flag;	
	int			auth_debug_flag;	
	apr_array_header_t *	rewrites;
	apr_array_header_t *	matches;
} auth_msfix_config_rec;                                // Configuration rec

/*
 * Macros
 */
#define strtrue(s) (s && *s)
#define	MATCH_EX	"^.*\\\\(.*)$"
#define REPLACE_EX	"$1"
#define DMSG( L )	if( s->auth_debug_flag ){ ap_log_rerror L ; }


/* ======================================================================== */
/*
 * Configuration Table: Define all of the commands availble in the config
 *
 */
static const command_rec auth_msfix_cmds[] =
{
    AP_INIT_FLAG(
	"AuthMSFix", 
	ap_set_flag_slot,
	(void *)APR_OFFSETOF(auth_msfix_config_rec, auth_msfix_flag ),
	OR_AUTHCFG,
	"On or Off depending on whether to cleanup after MS XP WebDav"),

    AP_INIT_FLAG(
	"AuthMSFixDebug", 
	ap_set_flag_slot,
	(void *)APR_OFFSETOF(auth_msfix_config_rec, auth_debug_flag ),
	OR_AUTHCFG,
	"On or Off depending on whether to print some debug info"),

    AP_INIT_TAKE2(
	"AuthMSFixMatch" ,
	add_msfix_rewrites,
	NULL,
	OR_AUTHCFG,
	"Enter the pattern to match and the replacement string on one line"
	 ),

    AP_INIT_TAKE2(
	"AuthMSFixOnHeader" ,
	add_msfix_onheader,
	NULL,
	OR_AUTHCFG,
	"Enter the header NAME to check and see if we have a MATCH"
	 ),

    {NULL}
};



/*
 * Register the routines we want called during which phases
 *
 */
static void my_register_hooks()
{
	/* We want to run before mod_auth runs. */
	/*static const char *aszpost[] = {"mod_auth.c", NULL};*/

	/* Authentication function */
	ap_hook_check_user_id(auth_msfix_handler,NULL,NULL,APR_HOOK_FIRST);

}

/*
 * Standard APACHE Declaration Table
 *
 */
module AP_MODULE_DECLARE_DATA auth_msfix_module = 
{
    STANDARD20_MODULE_STUFF, 
    auth_msfix_dir_config , /* dir config creator */
    NULL,                   /* merge  per-dir    config structures */
    NULL,                   /* create per-server config structures */
    NULL,                   /* merge  per-server config structures */
    auth_msfix_cmds,        /* [config file] command table */
    my_register_hooks       /* register hooks */
};


/*
 * Configuration Routine: Add entries to the REWRITE table
 * 	Referenced in the command_rec
 *
 */
static const char * add_msfix_rewrites( cmd_parms *cmd, void *dummy, const char *arg1, const char *arg2)
{
	rewrite_entry * r;
	auth_msfix_config_rec *d = dummy;

        if (!(d->rewrites )) {
	       d->rewrites = apr_array_make(cmd->pool, 2, sizeof( rewrite_entry ));
	}
	r = ( rewrite_entry *)apr_array_push(d->rewrites ) ;
	if( r ){
		r->match    = arg1;
		r->replace  = arg2;
		r->compiled = ap_pregcomp( cmd->pool, 	// Compile regex
			arg1,
			AP_REG_EXTENDED|AP_REG_ICASE );
		if( r->compiled == NULL )		// Did it compile? No?
		     return apr_psprintf( cmd->pool, "In MATCHES: unable to compile pattern %s", 
				     arg1 );
	}else{
		return "Could not add Matches entry";
	}

	return NULL;
}


/*
 * Configuration Routine: Add entries to the MATCHES table
 * 	Referenced in the command_rec
 *
 */
static const char * add_msfix_onheader( cmd_parms *cmd, void *dummy, const char *arg1, const char *arg2)
{
	matches_entry * r;
	auth_msfix_config_rec *d = dummy;

        if (!(d->matches)) {
	       d->matches= apr_array_make(cmd->pool, 2, sizeof( matches_entry ));
	}
	r = ( matches_entry *)apr_array_push(d->matches) ;
	if( r ){
		r->header  = arg1;
		r->match   = arg2;
		if( arg2[0] == '*' && arg2[1] == '\0' )
			r->compiled = NULL;
		else{
		r->compiled = ap_pregcomp( cmd->pool, 	// Compile regex
			arg2,
			AP_REG_EXTENDED|AP_REG_ICASE );
		if( r->compiled == NULL )		// Did it compile? No?
		     return apr_psprintf( cmd->pool, "In OnHeader: unable to compile pattern %s", 
				     arg2);
		}
	}else{
		return "Could not add OnHeader entry";
	}

	return NULL;
}



/*
 * Configuration Routine: Create the directory configuration
 * 	Referenced in the command_rec
 *
 */
static void *auth_msfix_dir_config(apr_pool_t *p, char *d)
{
    auth_msfix_config_rec *sec =
    (auth_msfix_config_rec *) apr_palloc(p, sizeof(auth_msfix_config_rec));
    
    sec->auth_msfix_flag	= 0;		/* Don't use module	*/
    sec->auth_debug_flag	= 0;		/* Don't use debug	*/
    sec->rewrites		= apr_array_make( p , 4 , sizeof( rewrite_entry ) );
    sec->matches 		= apr_array_make( p , 4 , sizeof( rewrite_entry ) );

    return sec;
}

/*
 * auth_msfix_handler
 * 	Called during the 'check_user' phase.
 *
 * 	This is performs all of the functions for patching/fixing the Auth header
 * 	It may affect the header but then ALWAYS returns 'declined'
 *
 */
static int auth_msfix_handler(request_rec *r)
{
	ap_regmatch_t pmatch[ 10 ];

	int i, found = 0;
	const char *t;
	const char *sent_pw;
	const char *check_header;
	char *user;
	const char *auth_line;
	matches_entry *mlst;
	rewrite_entry *rlst;

	char *AuthType = (PROXYREQ_PROXY == r->proxyreq)
			? "Proxy-Authorization" : "Authorization";


	auth_msfix_config_rec *s =
        (auth_msfix_config_rec *) ap_get_module_config(r->per_dir_config, &auth_msfix_module);           

	if (s->auth_msfix_flag != 1)		/* Not turned on? Skip */
		{ return DECLINED; }            

        rlst	= ( rewrite_entry *)(s->rewrites->elts);
	mlst = ( matches_entry *)(s->matches->elts);

	if( s->rewrites->nelts == 0 || s->matches->nelts == 0 )
	{ return DECLINED; }


	/*
	 * Because we have to change one of the headers (Auth Basic)
	 * we are going to fetch and check the line ourselves. If there
	 * are any errors we just skip this step and let the others 
	 * handle the processing.
	 *
	 * This section is lifted (borrowed!) from apache's protocol.c
	*/


	if (!(t = ap_auth_type(r)) || strcasecmp(t, "Basic")){
		return DECLINED;}			// Only BASIC auth

        auth_line	= apr_table_get(r->headers_in, AuthType );
	if (!auth_line) { return DECLINED; }		// No AUTH from client

	if (strcasecmp(ap_getword(r->pool, &auth_line, ' '), "Basic")) {
		return DECLINED; }			// Get Auth TYPE

	while (*auth_line == ' ' || *auth_line == '\t') {
		         auth_line++;			// Skip blanks
	}

	t = ap_pbase64decode(r->pool, auth_line);	// Decode text
	user = ap_getword_nulls (r->pool, &t, ':');	// Break apart
	sent_pw = t;					// User:Pwd


     if( user == NULL || *user == '\0' )
	{ return DECLINED; }				// No username
     
	// Well, we've got the header and it is the right type. Now
	// check to see if this is a line we should process...


     DMSG( (APLOG_MARK, APLOG_ERR, 0,r, "BEGIN AUTH_MSFIX _____________") );
     if( mlst[0].compiled == NULL ){
	     found++;					// Match all 
	     DMSG( (APLOG_MARK, APLOG_ERR, 0,r, 
			"No testing of headers (match all)") );
     }else						// Check headers
	for( i = 0; i< s->matches->nelts; i++)		// For all tests...
	{						// Fetch entries
             DMSG( (APLOG_MARK, APLOG_ERR, 0,r, 
			     "Test for HEADER '%s' REGEX '%s'", 
			     mlst[i].header, mlst[i].match ));
	     check_header = apr_table_get(r->headers_in, mlst[i].header );
	     if( check_header == NULL ){		// No hdr of type
	 	  DMSG( (APLOG_MARK, APLOG_ERR, 0,r, 
				"No Client Header of that type!") );
		     continue;				// .. skip
	     }
 	  DMSG( (APLOG_MARK, APLOG_ERR, 0,r, "Client Header is '%s'", 
				  check_header) );

     	   if( ap_regexec( mlst[i].compiled, check_header,10,pmatch,0) == 0 )
	   {
		   ++found;				// Found a match
	 	  DMSG( (APLOG_MARK, APLOG_ERR, 0,r, "Match") );
		   break;				// no more test
	   }
	} // end for each header


	if( ! found ){					// No header found?
		return DECLINED;			// .. go no further
	}
	/* ---------------------------------------------------------------
	 *		Fix Up AUTH Header from Client
	 * ---------------------------------------------------------------
	 */

	DMSG( (APLOG_MARK, APLOG_ERR, 0,r, "Test for AUTH headers of user '%s'", user ) );
	for( i = 0; i< s->rewrites->nelts; i++)		// For all auth rules
	{
             DMSG( (APLOG_MARK, APLOG_ERR, 0,r, 
			     "Try REGEX '%s' REPLACE '%s'", 
			     rlst[i].match, rlst[i].replace));

     	   if( ap_regexec( rlst[i].compiled, user, 10 , pmatch , 0 ) == 0 ){
	 	DMSG( (APLOG_MARK, APLOG_ERR, 0,r, "Match '%s'", user) );
	        user = ap_pregsub( r->pool, rlst[i].replace ,	// do replacement
			     	user, 10  , pmatch );
		if( user == NULL ){			// Bad replacement?
			ap_log_rerror( APLOG_MARK, APLOG_ERR, 0 , r ,
			     "Unable to perform replacement Match=%s, Replace=%s",
			     rlst[i].match , rlst[i].replace );
			continue;			// Skip replacement
		}
	 	DMSG( (APLOG_MARK, APLOG_ERR, 0,r, "User now '%s'", user) );
	        auth_line = apr_pstrcat( r->pool, 	// Create new header
			"Basic " , 			// for basic
			ap_pbase64encode( r->pool, 	// auth.
			     apr_pstrcat( r->pool, 	// (base 64 of
				     user, 		// USER:PWD
				     ":", 		
				     sent_pw , 	
				     NULL))
			, NULL);
	     apr_table_set( r->headers_in , AuthType , auth_line );
	     return( DECLINED );
		  
	  }else{
	 	DMSG( (APLOG_MARK, APLOG_ERR, 0,r, "No Match") );
	  } // endif regexec
	} // endfor all auth rules

	return( DECLINED );

} //end auth_msfix_handler

